#= Metasequoiaf[^DOM`
# 'mqo/dom'
require 'mrlib'

module Metasequoia
   class UV < MetaRed::UV
      def to_s
         sprintf("%.5f %.5f", @u, @v)
      end
   end

   class Vector < MetaRed::Vector
      # fnumŏ_ȉ̌w肷
      def to_s(fnum = 4)
         return sprintf('%.4f %.4f %.4f', @x, @y, @z) if fnum == 4
         return sprintf("%.#{fnum}f %.#{fnum}f %.#{fnum}f", @x, @y, @z)
      end
   end

   class Vertex < Vector
   end

   class Color < MetaRed::Vector
      attr_accessor :alpha
      def initialize(red, green, blue, alpha = nil)
         @x, @y, @z, @alpha = red, green, blue, alpha
      end
      alias r x
      alias g y
      alias b z
      def to_s
         return sprintf("%.3f %.3f %.3f", @x, @y, @z) if @alpha.nil?
         return sprintf("col(%.3f %.3f %.3f %.3f)", @x, @y, @z, @alpha)
      end
   end

   class Face
      attr_reader :vnum, :vertex_id, :uv, :vcol
      attr_accessor :material
      def to_s
         tmp = "#{@vnum} V(#{@vertex_id.join(' ')})"
         tmp += " M(#{@material})" if @material != :default
         tmp += " UV(#{@uv.join(' ')})" unless @uv.nil?
         tmp += " COL(#{@vcol.join(' ')})" unless @vcol.nil?
         return tmp
      end
   end

   class Material
      attr_accessor(
         :name, :color, :dif, :amb, :emi, :spc, :power, :tex, :aplane, :bump,
         :vcol, :shader, :proj_type, :proj_pos, :proj_scale, :proj_angle
      )
      def to_s
         tmp = "\t\"#{@name}\""
         tmp += " shader(#{@shader})" unless @shader.nil?
         tmp += " vcol(#{@vcol})" unless @vcol.nil?
         tmp += sprintf(" %s dif(%.3f) amb(%.3f) emi(%.3f) spc(%.3f) power(%.2f)",\
                  @color.to_s, @dif, @amb, @emi, @spc, @power)
         tmp += " tex(\"#{@tex}\")" unless @tex.nil?
         tmp += " aplane(\"#{@aplane}\")" unless @aplane.nil?
         tmp += " bump(\"#{@bump}\")" unless @bump.nil?
         tmp += " proj_type(#{@proj_type})" unless @proj_type.nil?
         tmp += " proj_pos(#{@proj_pos})" unless @proj_pos.nil?
         tmp += " proj_scale(#{@proj_scale})" unless @proj_scale.nil?
         tmp += " proj_angle(#{@proj_angle})" unless @proj_angle.nil?
         return tmp
      end
      def fputs(io = STDOUT)
         io.print self.to_s,"\n"
      end
   end

   class BackImage
      attr_reader :view, :file, :opt
      def to_s
         return "\t#{@view} \"#{@file}\"" if @opt.nil?
         "\t#{@view} \"#{@file}\" #{@opt.join(' ')}"
      end
      def fputs(io = STDOUT)
         io.print self.to_s,"\n"
      end
   end

   class VertexAttr
      attr_accessor :weit
      def to_s
         return '' if @weit.nil? || @weit.empty?
         "\tvertexattr {\n#{@weit}\t}\n"
      end
      def fputs(io = STDOUT)
         return if @weit.nil? || @weit.empty?
         io.print "\tvertexattr {\n"
         @weit.fputs(io)
         io.print "\t}\n"
      end
   end

   # Metasequoia::Object != Object
   class Object
      # colorVector^
      attr_accessor(
         :name, :visible, :locking, :shading, :facet,
         :color, :color_type, :vertices, :faces, :patch, :segment,
         :mirror, :mirror_axis, :mirror_dis,
         :lathe, :lathe_axis, :lathe_seg,
         :vertexattr
      )
      def to_s
         tmp = 'Object "'+@name+'" {'+"\n"
         tmp += "\tvisible #{@visible}\n"
         tmp += "\tlocking #{@locking}\n"
         tmp += "\tshading #{@shading}\n" unless @shading.nil?
         tmp += "\tpatch #{@patch}\n" unless @patch.nil?
         tmp += "\tsegment #{@segment}\n" unless @segment.nil?
         tmp += "\tmirror #{@mirror}\n" unless @mirror.nil?
         tmp += "\tmirror_axis #{@mirror_axis}\n" unless @mirror_axis.nil?
         tmp += sprintf("\tmirror_dis %.3f\n", @mirror_dis) unless @mirror_dis.nil?
         tmp += "\tlathe #{@lathe}\n" unless @lathe.nil?
         tmp += "\tlathe_axis #{@lathe_axis}\n" unless @lathe_axis.nil?
         tmp += "\tlathe_seg #{@lathe_seg}\n" unless @lathe_seg.nil?
         tmp += sprintf("\tfacet %.1f\n", @facet)
         tmp += "\tcolor #{@color}\n"
         tmp += "\tcolor_type #{@color_type}\n"
         tmp += @vertices.to_s
         tmp += @faces.to_s
         tmp += @vertexattr.to_s unless @vertexattr.nil?
         tmp += '}'+"\n"
         return tmp
      end
      def fputs(io)
         io.print "Object \"#{@name}\" {\n"
         io.print "\tvisible #{@visible}\n"
         io.print "\tlocking #{@locking}\n"
         io.print "\tshading #{@shading}\n" unless @shading.nil?
         io.print "\tpatch #{@patch}\n" unless @patch.nil?
         io.print "\tsegment #{@segment}\n" unless @segment.nil?
         io.print "\tmirror #{@mirror}\n" unless @mirror.nil?
         io.print "\tmirror_axis #{@mirror_axis}\n" unless @mirror_axis.nil?
         io.print sprintf("\tmirror_dis %.3f\n", @mirror_dis) unless @mirror_dis.nil?
         io.print "\tlathe #{@lathe}\n" unless @lathe.nil?
         io.print "\tlathe_axis #{@lathe_axis}\n" unless @lathe_axis.nil?
         io.print "\tlathe_seg #{@lathe_seg}\n" unless @lathe_seg.nil?
         io.print sprintf("\tfacet %.1f\n", @facet)
         io.print "\tcolor #{@color}\n"
         io.print "\tcolor_type #{@color_type}\n"
         @vertices.fputs(io)
         @faces.fputs(io)
         @vertexattr.fputs(io) unless @vertexattr.nil?
         io.print "}\n"
      end
   end

   # @pos, @lookatVector^A@ambColor^
   class Scene
      attr_accessor :pos, :lookat, :head, :pich, :ortho, :zoom2, :amb
      def to_s
         tmp = "Scene {\n"
         tmp += "\tpos #{@pos}\n"
         tmp += "\tlookat #{@lookat}\n"
         tmp += sprintf("\thead %.4f\n", @head)
         tmp += sprintf("\tpich %.4f\n", @pich)
         tmp += "\tortho #{@ortho}\n"
         tmp += sprintf("\tzoom2 %.4f\n", @zoom2)
         tmp += "\tamb #{@amb}\n"
         tmp += "}\n"
         return tmp
      end
      def fputs(io = STDOUT)
         io.print "Scene {\n"
         io.print "\tpos #{@pos}\n"
         io.print "\tlookat #{@lookat}\n"
         io.printf("\thead %.4f\n", @head)
         io.printf("\tpich %.4f\n", @pich)
         io.print "\tortho #{@ortho}\n"
         io.printf("\tzoom2 %.4f\n", @zoom2)
         io.print "\tamb #{@amb}\n"
         io.print "}\n"
      end

      # Jʒu, _, , J̏, p(c) Ԃ
      def camera_info
         wX = Vector.new(1,0,0)
         wY = Vector.new(0,1,0)
         wZ = Vector.new(0,0,1)
         q = wY.rot_Quaternion(-@head) * wX.rot_Quaternion(-@pich)

         # ^ZR̃JWĂ悭킩(^^;)
         # Ƃ肠ȊłxČł...
         dxy = Vector.new(@pos.x,@pos.y,0)
         clook = @lookat - ((q * dxy)/q).vector
         cZ = ((q * wZ)/q).vector
         cpos = cZ*@pos.z + clook
         # 
         eye = -cZ
         # 
         up = ((cZ * wY) * cZ)
         # p(c)
         fov = Math::atan(1000/(@zoom2*@pos.z)) * 360/Math::PI

         return cpos, clook, eye, up, fov
      end
   end

   class Document
      attr_accessor :scene, :materials, :objects, :backimages
      def to_s
         tmp = "Metasequoia Document\nFormat Text Ver 1.0\n\n"
         tmp += @scene.to_s unless @scene.nil?
         tmp += @backimages.to_s unless @backimages.nil?
         tmp += @materials.to_s unless @materials.nil?
         tmp += @objects.to_s unless @objects.nil?
         tmp += "Eof\n"
         return tmp
      end
      def fputs(io = STDOUT)
         io.print "Metasequoia Document\nFormat Text Ver 1.0\n\n"
         @scene.fputs(io) unless @scene.nil?
         @backimages.fputs(io) unless @backimages.nil?
         @materials.fputs(io) unless @materials.nil?
         @objects.fputs(io) unless @objects.nil?
         io.print "Eof\n"
      end
   end

   # `Chunk̋ʃNX
   # Metasequoia::Chunk̃TuNX"#{@content_class}Chunk"ƂOɂāA
   # {̂ł@chunk@content_class̔zƂB
   # ܂AMetasequoia::HogeChunk@content_class = Metasequoia::HogeB
   # AMetasequoia::Hoge݂ȂƂObjectƂB
   class Chunk
      attr_reader :chunk
      private :chunk
      def initialize(chunk = [])
         case chunk
         when Integer
            @chunk = Array.new(chunk)
         when Array
            @chunk = chunk
            begin
               @content_class = eval("#{self.class}".sub(/Chunk/,'')) || Object
            rescue NameError
               @content_class = Object
            end
         else
            raise 'invalid argument type'
         end
      end
      def +(other)
         @chunk.concat(other) if other.class == Array
         @chunk.concat(other.chunk) if other.class == self.class
         return self
      end
      def <<(data)
         @chunk << data if data.kind_of?(@content_class)
      end
      def size
         @chunk.size
      end
      def [](idx)
         @chunk[idx]
      end
      def []=(idx, data)
         @chunk[idx] = data if data.kind_of?(@content_class)
      end
      def each( &block )
         @chunk.each( &block )
      end
      def each_with_index( &block )
         @chunk.each_with_index( &block )
      end
      def map( &block )
         @chunk.map( &block )
      end
      def uniq!
         @chunk.uniq!
      end
      def clear
         @chunk.clear
      end
      def empty?
         @chunk.empty?
      end
   end

   class Weit < Chunk
      def initialize(chunk = [])
         @chunk = chunk
      end
      def <<(data)
         @chunk << data if data.is_a?(Array) && data.size == 2
      end
      def to_s
         return '' if @chunk.empty?
         ret = "\t\tweit {\n"
         @chunk.each{|x|
            ret += "\t\t\t#{x[0]} #{x[1]}\n"
         }
         ret += "\t\t}\n"
         return ret
      end
      def fputs(io = STDOUT)
         return if @chunk.empty?
         io.print "\t\tweit {\n"
         @chunk.each{|x|
            io.print "\t\t\t#{x[0]} #{x[1]}\n"
         }
         io.print "\t\t}\n"
      end
   end

   class BackImageChunk < Chunk
      def to_s
         return '' if @chunk.empty?
         tmp = "BackImage {\n"
         self.each{|b|
            tmp += "\t#{b}\n"
         }
         tmp += "}\n"
         return tmp
      end
      def fputs(io = STDOUT)
         return if @chunk.empty?

         io.print "BackImage {\n"
         self.each{|b| b.fputs(io) }
         io.print "}\n"
      end
      def [](key)
         self.each{|bg|
            return bg if bg.view == key
         }
         return nil
      end
   end

   class MaterialChunk < Chunk
      def to_s
         tmp = "Material #{self.size} {\n"
         self.each{|m|
            tmp += "\t"+m.to_s+"\n"
         }
         tmp += "}\n"
         return tmp
      end
      def fputs(io = STDOUT)
         io.print "Material #{self.size} {\n"
         self.each{|m| m.fputs(io) }
         io.print "}\n"
      end
   end

   class ObjectChunk < Chunk
      def to_s
         tmp = ''
         self.each{|o|
            tmp += o.to_s
         }
         return tmp
      end
      def fputs(io = STDOUT)
         self.each{|o|
            o.fputs(io)
         }
      end
   end

   class VertexChunk < Chunk
      attr_accessor :weit
      def to_s
         tmp = "\tvertex #{self.size} {\n"
         self.each{|v|
            tmp += "\t\t#{v}\n"
         }
         tmp += "\t}\n"
         return tmp
      end
      def fputs(io = STDOUT)
         io.print "\tvertex #{self.size} {\n"
         self.each{|v|
            io.printf("\t\t%s\n", v)
         }
         io.print "\t}\n"
      end
   end

   class FaceChunk < Chunk
      def to_s
         tmp = "\tface #{self.size} {\n"
         self.each{|f|
            tmp += "\t\t#{f}\n"
         }
         tmp += "\t}\n"
         return tmp
      end
      def fputs(io = STDOUT)
         io.print "\tface #{self.size} {\n"
         self.each{|f|
            io.printf("\t\t%s\n", f)
         }
         io.print "\t}\n"
      end
   end

end
