#= Ruby/MetaRed T|[gCu

module MetaRed
   class Vector
      attr_accessor :x, :y, :z
      def initialize(x, y, z)
         @x = x
         @y = y
         @z = z
      end

      def to_s
         return "(#{@x}, #{@y}, #{@z})"
      end

      def ==(other)
         return false if other.class != self.class
         return @x == other.x && @y == other.y && @z = other.z
      end

      def -@
         return self.class.new(-@x, -@y, -@z)
      end

      def -(other)
         raise 'Type mismatch' if other.class != self.class
         return self.class.new(@x - other.x, @y - other.y, @z - other.z)
      end

      def +(other)
         raise 'Type mismatch' if other.class != self.class
         return self.class.new(@x + other.x, @y + other.y, @z + other.z)
      end

      # O or XJ[{
      def *(other)
         case other
         when self.class
            return self.class.new(
               @y * other.z - @z * other.y,
               @z * other.x - @x * other.z,
               @x * other.y - @y * other.x
            )
         when Numeric
            return self.class.new(@x * other, @y * other, @z * other)
         else
            raise 'Type mismatch'
         end
      end

      # XJ[{
      def /(other)
         raise 'Type mismatch' unless other.kind_of?(Numeric)
         return self.class.new(@x / other, @y / other, @z / other)
      end

      # 
      def &(other)
         raise 'Type mismatch' if other.class != self.class
         return @x * other.x + @y * other.y + @z * other.z
      end

      # m(ς̕)
      def norm
         return Math.sqrt(@x**2 + @y**2 + @z**2)
      end

      # K
      def normalize
         n = self.norm
         return self.class.new(@x / n, @y / n, @z / n)
      end

      # K(jI)
      def normalize!
         n = self.norm
         @x /= n
         @y /= n
         @z /= n
         return self
      end

      # otherƂ̋
      # (self - other).norm ł܂邪A
      # self - otherłnew邩炻ꂪȏꍇɎgƂ悢
      def distance(other)
         raise 'Type mismatch' if other.class != self.class
         return Math.sqrt( (@x - other.x)**2 + (@y - other.y)**2 + (@z - other.z)**2 )
      end

      def to_Quaternion
         return Quaternion.new(0, self)
      end

      # selftheta](E)̃NEH[^jIԂB
      # theta̒Pʂ̓WAB
      def rot_Quaternion(theta)
         raise 'Type mismatch' unless theta.kind_of?(Numeric)
         v = self.normalize
         return Quaternion.new(Math.cos(theta/2), v * Math.sin(theta/2))
      end

      # otherselftheta](E)xNgԂB
      # theta̒Pʂ̓WAB
      def rot(theta, other)
         raise 'Type mismatch of 1st arg.' unless theta.kind_of?(Numeric)
         raise 'Type mismatch of 2nd arg.' if other.class != self.class
         q = self.rot_Quaternion(theta)
         v = (q * other.to_Quaternion) / q
         return v.vector
      end
   end

   class Quaternion
      attr_accessor :scalar, :vector
      def initialize(scalar, vector)
         raise 'Type mismatch of 1st arg.' unless scalar.kind_of?(Numeric)
         raise 'Type mismatch of 2nd arg.' unless vector.kind_of?(Vector)
         @scalar = scalar
         @vector = vector
      end

      def +(other)
         raise 'Type mismatch' unless other.kind_of?(Quaternion)
         return Quaternion.new(@scalar + other.scalar, @vector + other.vector)
      end

      def -(other)
         raise 'Type mismatch' unless other.kind_of?(Quaternion)
         return Quaternion.new(@scalar - other.scalar, @vector - other.vector)
      end

      def *(other)
         case other
         when Quaternion
            return Quaternion.new(
               @scalar * other.scalar - (@vector & other.vector),
               other.vector * @scalar + @vector * other.scalar + @vector * other.vector
            )
         when Vector
            return self * other.to_Quaternion
         when Numeric
            return Quaternion.new(@scalar * other, @vector * other)
         else
            raise 'Type mismatch'
         end
      end

      # 
      def c
         return Quaternion.new(@scalar, - @vector)
      end

      # m
      def norm
         #return (self * self.c).scalar
         return @scalar**2 + @vector.x**2 + @vector.y**2 + @vector.z**2
      end

      # 傫(m̕)
      def size
         return Math.sqrt(self.norm)
      end

      # self^(-1)
      def inv
         #return self.c / self.norm
         n = self.norm
         raise 'Not exists inverse' if n == 0
         return Quaternion.new(@scalar / n, @vector / (-n))
      end

      def /(other)
         case other
         when Quaternion
            return self * other.inv
         when Numeric
            return Quaternion.new(@scalar / other, @vector / other)
         else
            raise 'Type mismatch'
         end
      end
   end

   class UV
      attr_accessor :u, :v
      def initialize(u, v)
         @u, @v = u, v
      end

      def to_s
         return "(#{@u}, #{@v})"
      end

      def ==(other)
         return false if other.class != self.class
         return @u == other.u && @v == other.v
      end

      def -@
         return self.class.new(-@u, -@v)
      end

      def -(other)
         raise 'Type mismatch' if other.class != self.class
         return self.class.new(@u - other.u, @v - other.v)
      end

      def +(other)
         raise 'Type mismatch' if other.class != self.class
         return self.class.new(@u + other.u, @v + other.v)
      end

      # XJ[{
      def /(other)
         raise 'Type mismatch' unless other.kind_of?(Numeric)
         return self.class.new(@u / other, @v / other)
      end

      # 
      def &(other)
         raise 'Type mismatch' if other.class != self.class
         return @u * other.u + @v * other.v
      end

      # m(ς̕)
      def norm
         return Math.sqrt(@u**2 + @v**2)
      end

      # K
      def normalize
         n = self.norm
         return self.class.new(@u / n, @v / n)
      end

      # K(jI)
      def normalize!
         n = self.norm
         @u /= n
         @v /= n
         return self
      end
   end
end
